
// This header file is auto-generated from the file cosmosis/datablock/section_names.txt.
// Edit that and then re-make to add your own pre-defined section names.

#define LIKELIHOODS_SECTION "likelihoods"
#define DATA_VECTOR_SECTION "data_vector"
#define COSMOLOGICAL_PARAMETERS_SECTION "cosmological_parameters"
#define HALO_MODEL_PARAMETERS_SECTION "halo_model_parameters"
#define BARYON_PARAMETERS_SECTION "baryon_parameters"
#define SHEAR_CALIBRATION_PARAMETERS_SECTION "shear_calibration_parameters"
#define NUMBER_DENSITY_PARAMS_SECTION "number_density_params"
#define POST_FRIEDMANN_PARAMETERS_SECTION "post_friedmann_parameters"
#define INTRINSIC_ALIGNMENT_PARAMETERS_SECTION "intrinsic_alignment_parameters"
#define INTRINSIC_POWER_SECTION "intrinsic_power"
#define GALAXY_INTRINSIC_POWER_SECTION "galaxy_intrinsic_power"
#define MATTER_INTRINSIC_POWER_SECTION "matter_intrinsic_power"
#define WL_NUMBER_DENSITY_SECTION "wl_number_density"
#define MATTER_POWER_LIN_SECTION "matter_power_lin"
#define MATTER_POWER_NL_SECTION "matter_power_nl"
#define DARK_MATTER_POWER_LIN_SECTION "dark_matter_power_lin"
#define DARK_MATTER_POWER_NL_SECTION "dark_matter_power_nl"
#define BARYON_POWER_LIN_SECTION "baryon_power_lin"
#define BARYON_POWER_NL_SECTION "baryon_power_nl"
#define PHOTON_POWER_LIN_SECTION "photon_power_lin"
#define PHOTON_POWER_NL_SECTION "photon_power_nl"
#define MASSLESS_NEUTRINO_POWER_LIN_SECTION "massless_neutrino_power_lin"
#define MASSLESS_NEUTRINO_POWER_NL_SECTION "massless_neutrino_power_nl"
#define MASSIVE_NEUTRINO_POWER_LIN_SECTION "massive_neutrino_power_lin"
#define MASSIVE_NEUTRINO_POWER_NL_SECTION "massive_neutrino_power_nl"
#define CDM_BARYON_POWER_LIN_SECTION "cdm_baryon_power_lin"
#define CDM_BARYON_POWER_NL_SECTION "cdm_baryon_power_nl"
#define MATTER_DE_POWER_LIN_SECTION "matter_de_power_lin"
#define MATTER_DE_POWER_NL_SECTION "matter_de_power_nl"
#define WEYL_CURVATURE_POWER_LIN_SECTION "weyl_curvature_power_lin"
#define WEYL_CURVATURE_POWER_NL_SECTION "weyl_curvature_power_nl"
#define CDM_VELOCITY_POWER_LIN_SECTION "cdm_velocity_power_lin"
#define CDM_VELOCITY_POWER_NL_SECTION "cdm_velocity_power_nl"
#define BARYON_VELOCITY_POWER_LIN_SECTION "baryon_velocity_power_lin"
#define BARYON_VELOCITY_POWER_NL_SECTION "baryon_velocity_power_nl"
#define BARYON_CDM_RELATIVE_VELOCITY_POWER_LIN_SECTION "baryon_cdm_relative_velocity_power_lin"
#define BARYON_CDM_RELATIVE_VELOCITY_POWER_NL_SECTION "baryon_cdm_relative_velocity_power_nl"
#define MATTER_GALAXY_POWER_SECTION "matter_galaxy_power"
#define GALAXY_POWER_SECTION "galaxy_power"
#define SHEAR_XI_SECTION "shear_xi"
#define SHEAR_CL_SECTION "shear_cl"
#define SHEAR_CL_GG_SECTION "shear_cl_gg"
#define SHEAR_CL_GI_SECTION "shear_cl_gi"
#define SHEAR_CL_II_SECTION "shear_cl_ii"
#define GALAXY_CL_SECTION "galaxy_cl"
#define CMB_CL_SECTION "cmb_cl"
#define LSS_AUTOCORRELATION_SECTION "lss_autocorrelation"
#define GAL_MATTER_POWER_LIN_SECTION "gal_matter_power_lin"
#define LINEAR_CDM_TRANSFER_SECTION "linear_cdm_transfer"
#define SIGMA_R_LIN_SECTION "sigma_r_lin"
#define SIGMA_R_NL_SECTION "sigma_r_nl"
#define GALAXY_XI_SECTION "galaxy_xi"
#define GALAXY_SHEAR_XI_SECTION "galaxy_shear_xi"
#define GALAXY_INTRINSIC_XI_SECTION "galaxy_intrinsic_xi"
#define PLANCK_SECTION "planck"
#define INTRINSIC_ALIGNMENT_FIELD_SECTION "intrinsic_alignment_field"
#define FGAS_PARAMETERS_SECTION "fgas_parameters"
#define SHEAR_CALIBRATION_SECTION "shear_calibration"
#define SUPERNOVA_PARAMS_SECTION "supernova_params"
#define BIAS_FIELD_SECTION "bias_field"
#define GALAXY_LUMINOSITY_FUNCTION_SECTION "galaxy_luminosity_function"
#define DISTANCES_SECTION "distances"
#define DE_EQUATION_OF_STATE_SECTION "de_equation_of_state"
#define GROWTH_PARAMETERS_SECTION "growth_parameters"
#define TEST_PARAMETERS_SECTION "test_parameters"
#define MASS_FUNCTION_SECTION "mass_function"
#define EVS_SECTION "evs"
#define CLUSTERS_SECTION "clusters"

